'''
This module contains the Results superclass and all corresponding subclasses.
'''

from __future__ import(
	division,
	print_function,
	)

__docformat__ = 'restructuredtext en'
__all__ = ['RpoIsotopes']

import matplotlib.pyplot as plt
import numpy as np
import warnings

from collections import Sequence

#import exceptions
from .exceptions import(
	ArrayError,
	LengthError,
	StringError,
	)

#import helper functions
from .core_functions import(
	assert_len,
	)

from .plotting_helper import(
	_plot_dicts_iso,
	_rem_dup_leg,
	)

from .summary_helper import(
	_calc_ri_info,
	)

from .results_helper import(
	_calc_E_frac,
	_rpo_blk_corr,
	_rpo_extract_iso,
	_rpo_kie_corr,
	_rpo_mass_bal_corr,
	)


class Results(object):
	'''
	Class to store resulting data (e.g. Rpo isotopes, pyGC composition, etc.).
	Intended for subclassing. Do not call directly.
	'''

	def __init__(self):
		'''
		Initialize the superclass
		'''
		raise NotImplementedError

	#define classmethod for creating instance and populating measured values
	# directly from a .csv file
	@classmethod
	def from_csv(cls, file):
		raise NotImplementedError

	#define plotting method
	def plot(self, ax = None, labs = None, md = None, rd = None):
		'''
		Method for plotting ``rp.TimeData`` instance data.

		Parameters
		----------
		axis : matplotlib.axis or None
			Axis handle to plot on. Defaults to `None`.

		labs : tuple
			Tuple of axis labels, in the form (x_label, y_label).
			Defaults to `None`.

		md : tuple or None
			Tuple of modelled (i.e. rate distributions for each fraction)
			data, in the form  (x_data, y_data). Defaults to `None`.
			Only exists for plots of fraction-specific rate data.

		rd : tuple
			Tuple of real (observed) data, in the form (x_data, y_data). 
			Defaults to `None`. If this is a scatter plot of isotope results,
			then tuple is in the form (x_data, y_data, x_std, y_std).

		Returns
		-------
		ax : matplotlib.axis
			Updated axis handle containing data.
		'''

		#create axis if necessary and label
		if ax is None:
			_, ax = plt.subplots(1, 1)

		#label axes if labels exist
		if labs is not None:
			ax.set_xlabel(labs[0])
			ax.set_ylabel(labs[1])

		#add fraction-specific rate data if it exists
		if md is not None:

			#loop through each fraction and plot
			for frac in md[1]:

				ax.fill_between(
					md[0],
					np.zeros(len(md[0])),
					frac,
					facecolor = [0.3,0.3,0.3],
					edgecolor = 'k',
					alpha = 0.3,
					label = r'$\Pi_{f}(E)$')

			#set limits
			ax.set_xlim([0.9*np.min(md[0]), 1.1*np.max(md[0])])
			ax.set_ylim([0.9*np.min(md[1]), 1.1*np.max(md[1])])

			#add real data (i.e. total rate data) if it exists
			if rd is not None:
				ax.plot(
					rd[0], 
					rd[1],
					linewidth = 2,
					color = 'k',
					label = r'$p(0,E)$')

				#(re)set limits
				ax.set_xlim([0.9*np.min(rd[0]), 1.1*np.max(rd[0])])
				ax.set_ylim([0.9*np.min(rd[1]), 1.1*np.max(rd[1])])

		else:

			#plot the isotope vs. rate scatter plots
			ax.errorbar(
				rd[0], 
				rd[1], 
				xerr = rd[2], 
				yerr = rd[3],
				marker = 'o',
				ecolor = 'k',
				markersize = 12,
				mec = 'k',
				mfc = 'w',
				elinewidth = 1,
				markeredgewidth = 1,
				capsize = 0,
				ls = 'none')

			#set x limits
			ax.set_xlim([0.9*np.min(rd[0]), 1.1*np.max(rd[0])])

			#determine y spread and set limits
			ymin = np.min(rd[1]) - 0.1*(np.max(rd[1]) - np.min(rd[1]))
			ymax = np.max(rd[1]) + 0.1*(np.max(rd[1]) - np.min(rd[1]))

			ax.set_ylim([ymin, ymax])

		#remove duplicate legend entries
		han_list, lab_list = _rem_dup_leg(ax)
		
		ax.legend(
			han_list,
			lab_list, 
			loc = 'best',
			frameon = False)

		#make tight layout
		plt.tight_layout()

		return ax


class RpoIsotopes(Results):
	__doc__='''
	Class for inputting Ramped PyrOx isotopes, calculating p(0,E) contained in
	each RPO fraction, correcting isotope values for blank contribution, mass
	balance, and kinetic fractionation (d13C only), and storing resulting data
	and statistics.

	Parameters
	----------
	blk_corr : boolean
		Boolean to determine if inputted isotope data have been blank
		corrected, defaults to `False`.

	d13C_raw : None or array-like
		Array of the raw d13C values (VPDB) of each measured fraction, length
		`nFrac`. Defaults to `None`.

	d13C_raw_std : None or array-like
		The standard deviation of `d13C_raw` with length `nFrac`. Defaults to
		zeros or `None` if `d13C_raw` is `None`.

	Fm_raw : None or array-like
		Array of the  raw Fm values of each measured fraction, length `nFrac`.
		Defaults to `None`.

	Fm_raw_std : None or array-like
		The standard deviation of `Fm_raw` with length `nFrac`. Defaults to
		zeros or `None` if `Fm_raw` is `None`.

	kie_corr : boolean
		Boolean to determine if inputted d13C data have been fractionation
		corrected, defaults to `False`.

	m_raw : None or array-like
		Array of the raw masses (ugC) of each measured fraction, length
		`nFrac`. Defaults to `None`.

	m_raw_std : None or array-like
		The standard deviation of `d13C_raw` with length `nFrac`. Defaults to
		zero or `None` if `m_raw` is `None`.

	mb_corr : boolean
		Boolean to determine if inputted d13C data have been mass-balance
		corrected, defaults to `False`.

	model : rp.Daem
		``rp.Daem`` instance associated with the inputted energy complex,
		used for calculating the fractional E distributions and for KIE
		d13C correction.

	ratedata : rp.EnergyComplex
		``rp.EnergyComplex`` instance containing p(0,E) distribution for the
		thermogram associated with inputted isotopes. Used for calculating
		the fractional E distributions and for KIE d13C correction.

	t_frac : None or array-like
		2d array of the initial and final times of each fraction, in seconds.
		Shape [`nFrac` x 2]. Defaults to `None`.

	Warnings
	--------
	UserWarning
		If using an an isothermal model type for an RPO run.

	UserWarning
		If using a non-energy complex ratedata type for an RPO run.

	Raises
	------
	ArrayError
		If `t_frac` is not array-like.

	ArrayError
		If `nE` is not the same in the ``rp.Model`` instance and the 
		``rp.RateData`` instance.

	Notes
	-----
	When inputting `t_frac`, a time of 0 (i.e. `t0`, the initial time) is
	defined as the first timepoint in the ``RpoThermogram`` instance. If time
	passed between the thermogram `t0` and the beginning of fraction 1 
	trapping (as is almost always the case), `t_frac` must be adjusted 
	accordingly. This is done automatically when importing from .csv (see
	``RpoIsotopes.from_csv``) documenatation for info.

	See Also
	--------
	Daem
		``Model`` subclass used to generate the transform for RPO
		data and translate between time- and E-space.

	EnergyComplex
		``RateData`` subclass for storing and analyzing RPO rate data.

	RpoThermogram
		``TimeData`` subclass containing the time and fraction remaining data
		used for the inversion.

	Examples
	--------
	Generating a bare-bones isotope result instance containing only arbitrary
	time and Fm data for a given energy complex instance, ec, and a given
	model instance, Daem::

		#import modules
		import rampedpyrox as rp
		import numpy as np

		#generate arbitrary data for 3 fractions
		t_frac = [[100, 200], [200, 300], [300, 1000]]
		t_frac = np.array(t_frac)

		Fm_raw = [1.0, 0.5, 0.0]

		#create instance
		ri = rp.RpoIsotopes(
			daem,
			ec,
			t_frac,
			Fm_raw = Fm_raw)

	Generating a isotope result instance using an RPO output .csv file and the
	``RpoIsotopes.from_csv`` class method::

		#import modules
		import rampedpyrox as rp

		#create path to data file
		file = 'path_to_folder_containing_data/isotope_data.csv'

		#create instance
		ri = rp.RpoThermogram.from_csv(
			file,
			model,
			ratedata,
			blk_corr = True,
			mass_err = 0.01,
			DE = 0.0018)

	This will automatically correct inputted isotopes for the inputted
	instrument blank carbon contribution using the `blk_corr` flag and will
	assumed a 1 percent uncertainty in mass measurements. Additionally, this
	will fractionation-correct d13C data (if they exist) using a KIE DE of
	1.8 J/mol. **NOTE:** See ``RpoIsotopes.from_csv`` documentation for
	instructions on getting the .csv file in the right format.

	Plotting resulting p(0,E) contained in each RPO fraction::

		#import additional modules
		import matplotlib.pyplot as plt

		#create figure
		fig, ax = plt.subplots(1,3)

		#plot p(0,E) distributions
		ax[0] = ri.plot(
			ax = ax[0], 
			plt_var = 'p0E')

	Plotting resulting isotope vs. E scatter plots::
	
		#plot d13C data
		ax[1] = ri.plot(
			ax = ax[1], 
			plt_var = 'd13C',
			plt_corr = True) #plotting corrected values

		#plot Fm data
		ax[2] = ri.plot(
			ax = ax[2], 
			plt_var = 'Fm',
			plt_corr = True) #plotting corrected values

	Printing a summary of the raw and corrected isotope values::

		#raw fraction information
		print(ri.ri_raw_info)

		#corrected fraction information
		print(ri.ri_corr_info)

	**Attributes**

	d13C_corr : np.ndarray
		Array of the d13C values (VPDB) of each measured fraction, corrected
		for any of: blank, mass-balance, KIE. Length `nFrac`.

	d13C_corr_std : np.ndarray
		The standard deviation of the d13C values (VPDB) of each measured 
		fraction, corrected for any of: blank, mass-balance, KIE. Length 
		`nFrac`.

	d13C_raw : np.ndarray
		Array of the raw d13C values (VPDB) of each measured fraction, 
		length `nFrac`.

	d13C_raw_std : np.ndarray
		The standard deviation of `d13C_raw` with length `nFrac`.

	E_frac : np.ndarray
		Array of the mean E value (kJ) contained in each measured fraction as
		calculated by the inverse model, length `nFrac`.

	E_frac_std : np.ndarray
		The standard deviation of E (kJ) contained in each measured fraction
		as calculated by the inverse model, length `nFrac`.

	Fm_corr : np.ndarray
		Array of the blank-corrected Fm values of each measured fraction, 
		length `nFrac`.

	Fm_corr_std : np.ndarray
		The standard deviation of `Fm_corr` with length `nFrac`.

	Fm_raw : np.ndarray
		Array of the raw Fm values of each measured fraction, length `nFrac`.

	Fm_raw_std : np.ndarray
		The standard deviation of `Fm_raw` with length `nFrac`.

	m_corr : np.ndarray
		Array of the blank-corrected masses (ugC) of each measured fraction,
		length `nFrac`.

	m_corr_std : np.ndarray
		The standard deviation of `m_corr` with length `nFrac`.

	m_raw : np.ndarray
		Array of the raw masses (ugC) of each measured fraction, length
		`nFrac`.

	m_raw_std : np.ndarray
		The standard deviation of `m_raw` with length `nFrac`.

	nFrac : int
		The number of measured fractions.

	ri_corr_info : pd.DataFrame
		Dataframe containing the inputted summary info, using corrected
		isotopes: 

			time (init. and final), \n
			E (mean and std.), \n
			mass (mean and std.), \n
			d13C (mean and std.), \n
			Fm (mean and std.) \n

	ri_raw_info : pd.DataFrame
		Dataframe containing the inputted summary info, using raw isotopes: 

			time (init. and final), \n
			E (mean and std.), \n
			mass (mean and std.), \n
			d13C (mean and std.), \n
			Fm (mean and std.) \n

	t_frac : np.ndarray
		2d array of the initial and final times of each fraction, in seconds.
		Shape [`nFrac` x 2].
	'''

	def __init__(
			self,
			model,
			ratedata,
			t_frac,
			d13C_raw = None, 
			d13C_raw_std = None, 
			Fm_raw = None,
			Fm_raw_std = None,
			m_raw = None,
			m_raw_std = None,
			blk_corr = False,
			mb_corr = False,
			kie_corr = False):

		#assert all lenghts are the same and of the same length as t_frac, and
		# store as attributes
		if isinstance(t_frac, str):
			raise ArrayError(
				't_frac cannot be a string')

		elif isinstance(t_frac, Sequence) or hasattr(t_frac, '__array__'):
			
			n = len(t_frac)
			self.t_frac = t_frac
			self.nFrac = n

		else:
			raise ArrayError(
				't_frac must be array-like')

		#store existing data
		if m_raw is not None:
			self.m_raw = assert_len(m_raw, n)

			#store stdev if it exists, zeros if not
			if m_raw_std is not None:
				self.m_raw_std = assert_len(m_raw_std, n)
			else:
				self.m_raw_std = assert_len(0, n)

		if d13C_raw is not None:
			self.d13C_raw = assert_len(d13C_raw, n)

			#store stdev if it exists, zeros if not
			if d13C_raw_std is not None:
				self.d13C_raw_std = assert_len(d13C_raw_std, n)
			else:
				self.d13C_raw_std = assert_len(0, n)

		if Fm_raw is not None:
			self.Fm_raw = assert_len(Fm_raw, n)

			#store stdev if it exists, zeros if not
			if Fm_raw_std is not None:
				self.Fm_raw_std = assert_len(Fm_raw_std, n)
			else:
				self.Fm_raw_std = assert_len(0, n)

		#warn if model is not Daem
		mod_type = type(model).__name__

		if mod_type not in ['Daem']:
			warnings.warn(
				'Attempting to calculate E distributions using a model' 
				' instance of type %r. Consider using rp.Daem instance'
				' instead' % mod_type, UserWarning)

		#warn if ratedata is not EnergyComplex
		rd_type = type(ratedata).__name__

		if rd_type not in ['EnergyComplex']:
			warnings.warn(
				'Attempting to calculate E distributions using a ratedata'
				' instance of type %r. Consider using rp.EnergyComplex'
				' instance instead' % rd_type, UserWarning)

		#raise exception if not the right shape
		if model.nE != ratedata.nE:
			raise ArrayError(
				'Cannot combine model with nE = %r and RateData with'
				' nE = %r. Check that RateData was not created using a'
				' different model' % (model.nE, ratedata.nE))

		#calculate E distributions for each fraction
		E_frac, E_frac_std, p_frac = _calc_E_frac(self, model, ratedata)
		
		#store results
		self.E_frac = E_frac
		self.E_frac_std = E_frac_std

		#store protected energetic attributes
		self._p_frac = p_frac
		self._p = ratedata.p
		self._E = ratedata.E

		#store raw info summary table
		self.ri_raw_info = _calc_ri_info(self, flag = 'raw')

		#store bookkeeping corrected flags
		self._blk_corr = blk_corr
		self._mb_corr = mb_corr
		self._kie_corr = kie_corr

	#define classmethod for creating instance and populating measured values
	# directly from a .csv file
	@classmethod
	def from_csv(
			cls, 
			file,
			model,
			ratedata,
			blk_corr = False,
			blk_d13C = (-29.0, 0.1),
			blk_flux = (0.375, 0.0583),
			blk_Fm =  (0.555, 0.042),
			bulk_d13C_true = None,
			DE = 0.0018,
			mass_err = 0.01):
		'''
		Class method to directly import RPO fraction data from a .csv file and
		create an ``RpoIsotopes`` class instance.
		
		Parameters
		----------
		blk_corr : Boolean
			Tells the method whether or not to blank-correct isotope data. If
			`True`, blank-corrects according to inputted blank composition 
			values. If `bulk_d13C_true` is not `None`, further corrects d13C 
			values to ensure isotope mass balance (see Hemingway et al.,
			Radiocarbon **2017** for details).

		blk_d13C : tuple
			Tuple of the blank d13C composition (VPDB), in the form 
			(mean, stdev.) to be used of ``blk_corr = True``. Defaults to the
			NOSAMS RPO blank as calculated by Hemingway et al., Radiocarbon
			**2017**.

		blk_flux : tuple
			Tuple of the blank flux (ng/s), in the form (mean, stdev.) to
			be used of ``blk_corr = True``. Defaults to the NOSAMS RPO blank 
			as calculated by Hemingway et al., Radiocarbon **2017**.

		blk_Fm : tuple
			Tuple of the blank Fm value, in the form (mean, stdev.) to
			be used of ``blk_corr = True``. Defaults to the NOSAMS RPO blank 
			as calculated by Hemingway et al., Radiocarbon **2017**.

		bulk_d13C_true : None or array
			True measured d13C value (VPDB) for bulk material as measured
			independently (e.g. on a EA-IRMS). If not `None`, this value is
			used to mass-balance-correct d13C values as described in Hemingway
			et al., Radiocarbon **2017**. If not `none`, must be inputted in
			the form [mean, stdev.]

		DE : scalar
			Value for the difference in E between 12C- and 13C-containing
			atoms, in kJ. Defaults to 0.0018 (the best-fit value calculated
			in Hemingway et al., Radiocarbon, **2017**).

		file : str or pd.DataFrame
			File containing RPO isotope data, either as a string pointing
			to a .csv file or as a ``pd.DataFrame`` instance.

		mass_err : float
			Relative uncertainty in mass measurements, typically as a sum of
			manometric uncertainty in pressure measurements and uncertainty in
			vacuum line volumes. Defaults to 0.01 (i.e. 1\% relative 
			uncertainty).

		model : rp.Model
			``rp.Model`` instance containing the A matrix to use for 
			inversion.

		ratedata : rp.RateData
			``rp.Ratedata`` instance containing the reactive continuum data.
		
		Notes
		-----
		For bookkeeping purposes, the first 2 rows must be fractions "-1" and
		"0", where the timestamp for fraction "-1" is the first point in 
		the `all_data` file used to create the ``rp.RpoThermogram`` instance,
		and the timestamp for fraction "0" is the t0 for the first fraction.

		If mass, d13C, and Fm data exist, column names must be the following:

			'ug_frac' and 'ug_frac_std' \n
			'd13C' and 'd13C_std' \n
			'Fm' and 'Fm_std'

		See Also
		--------
		RpoThermogram.from_csv
			Classmethod for creating ``rp.RpoThermogram`` instance directly 
			from a .csv file.

		References
		----------
		[1] J.D. Hemingway et al. (2017) Assessing the blank carbon
			contribution, isotope mass balance, and kinetic isotope 
			fractionation of the ramped pyrolysis/oxidation instrument at 
			NOSAMS. *Radiocarbon*, **59**, 179-193.
		'''

		#extract data from file
		d13C, d13C_std, Fm, Fm_std, m, m_std, t_frac = _rpo_extract_iso(
			file,
			mass_err)

		#create RpoIsotopes instance and store raw data
		ri = cls(
			model,
			ratedata,
			t_frac,
			d13C_raw = d13C,
			d13C_raw_std = d13C_std,
			Fm_raw = Fm,
			Fm_raw_std = Fm_std,
			m_raw = m,
			m_raw_std = m_std,
			blk_corr = False,
			mb_corr = False,
			kie_corr = False)

		#blank correct m, d13C, Fm if necessary
		if blk_corr is True:

			#call blank- and mass-balance correction method
			ri.blank_correct(
				blk_d13C = blk_d13C,
				blk_flux = blk_flux,
				blk_Fm =  blk_Fm,
				bulk_d13C_true = bulk_d13C_true)

		#kinetic fractionation correct if necessary
		if DE is not None:

			#call kie correction method
			ri.kie_correct(
				model,
				ratedata,
				DE = DE)

		return ri

	def blank_correct(
		self,
		blk_d13C = (-29.0, 0.1),
		blk_flux = (0.375, 0.0583),
		blk_Fm =  (0.555, 0.042),
		bulk_d13C_true = None):
		'''
		Method to blank- and mass-balance correct raw isotope values.

		Parameters
		----------
		blk_d13C : tuple
			Tuple of the blank d13C composition (VPDB), in the form 
			(mean, stdev.) to be used of ``blk_corr = True``. Defaults to the
			NOSAMS RPO blank as calculated by Hemingway et al., Radiocarbon
			**2017**.

		blk_flux : tuple
			Tuple of the blank flux (ng/s), in the form (mean, stdev.) to
			be used of ``blk_corr = True``. Defaults to the NOSAMS RPO blank 
			as calculated by Hemingway et al., Radiocarbon **2017**.

		blk_Fm : tuple
			Tuple of the blank Fm value, in the form (mean, stdev.) to
			be used of ``blk_corr = True``. Defaults to the NOSAMS RPO blank 
			as calculated by Hemingway et al., Radiocarbon **2017**.

		bulk_d13C_true : None or array
			True measured d13C value (VPDB) for bulk material as measured
			independently (e.g. on a EA-IRMS). If not `None`, this value is
			used to mass-balance-correct d13C values as described in Hemingway
			et al., Radiocarbon **2017**. If not `none`, must be inputted in
			the form [mean, stdev.]

		Warnings
		--------
		UserWarning
			If already corrected for blank contribution
		
		UserWarning
			If already corrected for 13C mass balance

		References
		----------
		[1] J.D. Hemingway et al. (2017) Assessing the blank carbon
			contribution, isotope mass balance, and kinetic isotope 
			fractionation of the ramped pyrolysis/oxidation instrument at 
			NOSAMS. *Radiocarbon*, **59**, 179-193.
		'''

		#raise warnings
		if self._blk_corr == True:
			warnings.warn(
				'd13C, Fm, and/or m has already been blank-corrected!'
				' Proceeding anyway', UserWarning)

		if self._mb_corr == True:
			warnings.warn(
				'd13C has already been mass-balance corrected!'
				' Proceeding anyway', UserWarning)

		#define constants
		n = self.nFrac

		#extract d13C from self to be corrected
		if hasattr(self, 'd13C_corr'):
			d13C = self.d13C_corr
			d13C_std = self.d13C_corr_std
		
		else:
			d13C = self.d13C_raw
			d13C_std = self.d13C_raw_std

		#extract Fm from self to be corrected
		if hasattr(self, 'Fm_corr'):
			Fm = self.Fm_corr
			Fm_std = self.Fm_corr_std
		
		else:
			Fm = self.Fm_raw
			Fm_std = self.Fm_raw_std

		#extract m from self to be corrected
		if hasattr(self, 'm_corr'):
			m = self.m_corr
			m_std = self.m_corr_std
		
		else:
			m = self.m_raw
			m_std = self.m_raw_std

		#blank-correct values
		d13C, d13C_std, Fm, Fm_std, m, m_std = _rpo_blk_corr(
			d13C,
			d13C_std,
			Fm,
			Fm_std,
			m,
			m_std,
			self.t_frac,
			blk_d13C = blk_d13C,
			blk_flux = blk_flux,
			blk_Fm = blk_Fm)

		#set bookkeeping flag
		self._blk_corr = True

		#mass-balance correct d13C if necessary
		if bulk_d13C_true is not None:

			d13C, d13C_std = _rpo_mass_bal_corr(
				d13C,
				d13C_std,
				m,
				m_std,
				bulk_d13C_true)

			#set bookkeeping flag
			self._mb_corr = True

		#store corrected values if they exist
		if m is not None:
			self.m_corr = assert_len(m, n)

			#store stdev if it exists, zeros if not
			if m_std is not None:
				self.m_corr_std = assert_len(m_std, n)
			else:
				self.m_corr_std = assert_len(0, n)

		if d13C is not None:
			self.d13C_corr = assert_len(d13C, n)

			#store stdev if it exists, zeros if not
			if d13C_std is not None:
				self.d13C_corr_std = assert_len(d13C_std, n)
			else:
				self.d13C_corr_std = assert_len(0, n)

		if Fm is not None:
			self.Fm_corr = assert_len(Fm, n)

			#store stdev if it exists, zeros if not
			if Fm_std is not None:
				self.Fm_corr_std = assert_len(Fm_std, n)
			else:
				self.Fm_corr_std = assert_len(0, n)

		#generate summary table
		self.ri_corr_info = _calc_ri_info(self, flag = 'corr')

	def kie_correct(
		self,
		model,
		ratedata,
		DE = 0.0018):
		'''
		Method for further correcting d13C values to account for kinetic 
		isotope fractionation occurring within the instrument.

		Parameters
		----------
		model : rp.Model
			``rp.Model`` instance containing the A matrix to use for 
			inversion.

		ratedata : rp.RateData
			``rp.Ratedata`` instance containing the reactive continuum data.

		DE : scalar
			Value for the difference in E between 12C- and 13C-containing
			atoms, in kJ. Defaults to 0.0018 (the best-fit value calculated
			in Hemingway et al., Radiocarbon, **2017**).

		Warnings
		--------
		UserWarning
			If already corrected for kinetic fractionation

		References
		----------
		[1] J.D. Hemingway et al. (2017) Assessing the blank carbon
			contribution, isotope mass balance, and kinetic isotope 
			fractionation of the ramped pyrolysis/oxidation instrument at 
			NOSAMS. *Radiocarbon*, **59**, 179-193.
		'''

		#raise warnings
		if self._kie_corr == True:
			warnings.warn(
				'd13C has already been corrected for kinetic fractionation!'
				' Proceeding anyway', UserWarning)

		#set constants
		n = self.nFrac

		#extract d13C from self to be corrected
		if hasattr(self, 'd13C_corr'):
			d13C = self.d13C_corr
			d13C_std = self.d13C_corr_std
		
		elif hasattr(self, 'd13C_raw'):
			d13C = self.d13C_raw
			d13C_std = self.d13C_raw_std

		else:
			return

		#kie-correct values
		d13C, d13C_std = _rpo_kie_corr(
			self,
			d13C,
			d13C_std,
			model,
			ratedata,
			DE = DE)

		#set bookkeeping flag
		self._kie_corr = True

		#store results
		self.d13C_corr = assert_len(d13C, n)

		#store stdev if it exists, zeros if not
		if d13C_std is not None:
			self.d13C_corr_std = assert_len(d13C_std, n)
		else:
			self.d13C_corr_std = assert_len(0, n)

		#store summary
		self.ri_corr_info = _calc_ri_info(self, flag = 'corr')

	#define plotting method
	def plot(self, ax = None, plt_var = 'p0E', plt_corr = True):
		'''
		Method for plotting results, either p(0,E) distributions contained
		within each RPO fraction or isotopes vs. mean E for each RPO fraction.

		Parameters
		----------
		ax : None or matplotlib.axis
			Axis to plot on. If `None`, automatically creates a
			``matplotlip.axis`` instance to return. Defaults to `None`.

		plt_var : str
			Tells the method which variable to plot, available options are:
			'p0E' (for fraction-specific p(0,E) distributions), 'Fm', and d13C
			(isotope vs. fraction E scatter plots).

		plt_corr : str
			If `plt_var` is 'Fm' or 'd13C', `plt_corr` tells the method
			whether to plot raw or corrected values (if corrected values
			exist).

		Returns
		-------
		ax : matplotlib.axis
			Updated axis instance with plotted data.

		Raises
		------
		ArrayError
			if `plt_corr` is `True` but no corrected data exist.
			
		StringError
			If `plt_var` is not 'p0E', 'Fm', or 'd13C'.

		'''

		#check that plt_var is an appropriate string
		if plt_var not in ['p0E','Fm','d13C']:
			raise StringError(
				'plt_var does not accept %r. Must be "p0E", "Fm", or "d13C"'
				%plt_var)

		#check that corrected data exist if called
		if plt_var in ['Fm','d13C']:
			att = plt_var + '_corr'
			
			if plt_corr is True and not hasattr(self, att):
				raise ArrayError(
					'plt_corr is set to True but corrected values do not exist!')

		#extract axis label ditionary
		rpo_labs = _plot_dicts_iso('rpo_iso_labs', self)
		labs = (
			rpo_labs['E'][plt_var][0],
			rpo_labs['E'][plt_var][1])

		#determine attributes being plotted and generate dict
		if plt_var in ['Fm','d13C']:

			if plt_corr:
				att = plt_var + '_corr'

				#extract the right dict
				iso_rd = _plot_dicts_iso('iso_corr', self)
				rd = (
					iso_rd['E'][att][0],
					iso_rd['E'][att][1],
					iso_rd['E'][att][2],
					iso_rd['E'][att][3])

			else:
				att = plt_var + '_raw'

				#extract the right dict
				iso_rd = _plot_dicts_iso('iso_raw', self)
				rd = (
					iso_rd['E'][att][0],
					iso_rd['E'][att][1],
					iso_rd['E'][att][2],
					iso_rd['E'][att][3])

			#no modeled data exist
			md = None

		else:

			#generate rd and md for p(0,E) plots
			rd = (self._E, self._p)
			md = (self._E, self._p_frac)

		#call superclass method
		ax = super(RpoIsotopes, self).plot(
			ax = ax, 
			labs = labs,
			md = md,
			rd = rd)

		return ax

if __name__ == '__main__':

	import rampedpyrox as rp




from setuptools import setup

def readme():
	with open('README.rst') as f:
		return f.read()

setup(name='rampedpyrox',
	version='1.0.4',
	description='Ramped PyrOx kinetic and isotope analysis',
	long_description=readme(),
	classifiers=[
		'Development Status :: 3 - Alpha',
		'Intended Audience :: Science/Research',
		'License :: Free for non-commercial use',
		'Programming Language :: Python',
		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 2',
		'Programming Language :: Python :: 3.5',
		'Programming Language :: Python :: 2.7',
		'Topic :: Scientific/Engineering'
	],
	url='https://github.com/FluvialSeds/rampedpyrox',
	download_url='https://github.com/FluvialSeds/rampedpyrox/tarball/1.0.4',
	# download_url='https://github.com/FluvialSeds/rampedpyrox/archive/1.0.4.tar.gz',
	keywords=[
		'geochemistry',
		'isotopes',
		'kinetics',
		'inverse modeling',
		'carbon cycle'
	],
	author='Jordon D. Hemingway',
	author_email='jordon.hemingway@erdw.ethz.ch',
	license='GNU GPL Version 3',
	packages=['rampedpyrox'],
	install_requires=[
		'matplotlib',
		'numpy',
		'pandas',
		'scipy'
	],
	test_suite='nose.collector',
	tests_require=['nose'],
	include_package_data=True,
	zip_safe=False)


'''
Module to store all the core functions for ``rampedpyrox``.
'''

from __future__ import(
	division,
	print_function,
	)

__docformat__ = 'restructuredtext en'
__all__ = ['assert_len', 'calc_L_curve', 'derivatize', 'extract_moments']

import matplotlib.pyplot as plt
import numpy as np

from collections import Sequence

#import exceptions
from .exceptions import(
	ArrayError,
	LengthError,
	)

#define function to assert length of array
def assert_len(data, n):
	'''
	Asserts that an array has length `n` and `float` datatypes.

	Parameters
	----------
	data : scalar or array-like
		Array to assert has length n. If scalar, generates an np.ndarray
		with length n.

	n : int
		Length to assert

	Returns
	-------
	array : np.ndarray
		Updated array, now of class np.ndarray and with length n.

	Raises
	------
	ArrayError
		If inputted data not int or array-like (excluding string).

	LengthError
		If length of the array is not n.
	'''

	#assert that n is int
	n = int(n)

	#assert data is in the right form
	if isinstance(data, (int, float)):
		data = data*np.ones(n)
	
	elif isinstance(data, Sequence) or hasattr(data, '__array__'):
		
		if isinstance(data, str):
			raise ArrayError(
				'Data cannot be a string')

		elif len(data) != n:
			raise LengthError(
				'Cannot create array of length %r if n = %r' \
				% (len(data), n))

	else:
		raise ArrayError('data must be scalar or array-like')

	return np.array(data).astype(float)

#define package-level function for calculating L curves
def calc_L_curve(
	model, 
	timedata, 
	ax = None, 
	plot = False, 
	nLam = 150, 
	lam_max = 1e2, 
	lam_min = 1e-3):
	'''
	Function to calculate the L-curve for a given model and timedata
	instance in order to choose the best-fit smoothing parameter, `lambda`.

	Parameters
	----------
	model : rp.Model
		``rp.Model`` instance containing the A matrix to use for L curve 
		calculation.

	timedata : rp.TimeData
		``rp.TimeData`` instance containing the time and fraction remaining
		arrays to use in L curve calculation.

	Keyword Arguments
	-----------------
	ax : None or matplotlib.axis
		Axis to plot on. If `None` and ``plot = True``, automatically 
		creates a ``matplotlip.axis`` instance to return. Defaults to 
		`None`.

	plot : Boolean
		Tells the method to plot the resulting L curve or not.

	lam_min : int
		Minimum lambda value to search. Defaults to 1e-3.

	lam_max : int
		Maximum lambda value to search. Defaults to 1e2.

	nLam : int
		Number of lambda values to consider. Defaults to 150.

	Returns
	-------
	lam_best : float
		The calculated best-fit lambda value.
		
	axis : None or matplotlib.axis
		If ``plot = True``, returns an updated axis handle with plot.
	
	Raises
	------
	ScalarError
		If `lam_max` or `lam_min` are not int or float.

	ScalarError
		If `nLam` is not int.

	See Also
	--------
	Daem.calc_L_curve
		Instance method for ``calc_L_curve``.

	References
	----------
	[1] D.C. Forney and D.H. Rothman (2012) Inverse method for calculating
		respiration rates from decay time series. *Biogeosciences*, **9**,
		3601-3612.

	[2] P.C. Hansen (1994) Regularization tools: A Matlab package for analysis
		and solution of discrete ill-posed problems. *Numerical Algorithms*, 
		**6**, 1-35.
	'''

	a = model.calc_L_curve(
		timedata, 
		ax = ax, 
		plot = plot, 
		nLam = nLam, 
		lam_max = lam_max, 
		lam_min = lam_max)

	return a

#define function to derivatize an array w.r.t. another array
def derivatize(num, denom):
	'''
	Method for derivatizing numerator, `num`, with respect to denominator, 
	`denom`.

	Parameters
	----------
	num : int or array-like
		The numerator of the numerical derivative function.

	denom : array-like
		The denominator of the numerical derivative function. Length `n`.

	Returns
	-------
	derivative : rparray
		An ``np.ndarray`` instance of the derivative. Length `n`.

	Raises
	------
	ArrayError
		If `denom` is not array-like.

	See Also
	--------
	numpy.gradient
		The method used to calculate derivatives

	Notes
	-----
	This method uses the ``np.gradient`` method to calculate derivatives. If
	`denom` is a scalar, resulting array will be all ``np.inf``. If both `num`
	and `denom` are scalars, resulting array will be all ``np.nan``. If 
	either `num` or `denom` are 1d and the other is 2d, derivative will be
	calculated column-wise. If both are 2d, each column will be derivatized 
	separately.
	'''

	#assert denom is the right type
	if isinstance(denom, Sequence) or hasattr(denom, '__array__'):
			if isinstance(denom, str):
				raise ArrayError('denom cannot be a string')

	else:
		raise ArrayError('denom must be array-like')

	#make sure the arrays are the same length, or convert num to array if 
	# scalar
	n = len(denom)
	num = assert_len(num, n)

	#calculate separately for each dimensionality case
	if num.ndim == denom.ndim == 1:
		dndd = np.gradient(num)/np.gradient(denom)

	elif num.ndim == denom.ndim == 2:
		dndd = np.gradient(num)[0]/np.gradient(denom)[0]

	#note recursive list comprehension when dimensions are different
	elif num.ndim == 2 and denom.ndim == 1:
		col_der = [derivatize(col, denom) for col in num.T]
		dndd = np.column_stack(col_der)

	elif num.ndim == 1 and denom.ndim == 2:
		col_der = [derivatize(num, col) for col in denom.T]
		dndd = np.column_stack(col_der)

	return dndd

#define function to extract 1st and 2nd moments for a distribution
def extract_moments(x, y):
	'''
	Extracts 1st (mean) and 2nd (stdev) moments from a distribution.

	Parameters
	----------
	x : np.ndarray
		Array of x values, length `n`.

	y : np.ndarray
		Array of y values, length `n`.

	Returns
	-------
	mu : float
		First moment of distribution.

	sigma : float
		Second moment of distribution.
	'''

	#assert lengths
	n = len(x)
	y = assert_len(y, n)

	#calculate first moment
	scalar = 1/np.sum(y*np.gradient(x))

	mu = np.sum(x*y*scalar*np.gradient(x))
	sigsq = np.sum((x - mu)**2 * y*scalar*np.gradient(x))
	sigma = sigsq**0.5

	return mu, sigma

#define function for plotting raw thermogram with isotopes
def plot_tg_isotopes(timedata, result, ax = None, plt_corr = True):
	'''
	Function to plot raw timedata (e.g. RPO thermogram) and isotope values.

	Parameters
	----------
	ax : None or matplotlib.axis
		Axis to plot on. If `None`, automatically creates a
		``matplotlip.axis`` instance to return. Defaults to `None`.
	
	plt_corr : str
		If `plt_var` is 'Fm' or 'd13C', `plt_corr` tells the method
		whether to plot raw or corrected values (if corrected values
		exist).

	result : rp.Results
		``rp.Results`` instance containing the isotope results to plot.

	timedata : rp.TimeData
		``rp.TimeData`` instance containing the derivative timedata (e.g. rpo
		 thermogram) to plot.

	Returns
	-------
	ax : matplotlib.axis
		Updated axis instance with plotted data.

	Warnings
	--------
	UserWarning
		If `timedata` does not contain derivative timedata, dgdt.
	
	UserWarning
		If `result` does not contain any of the necessary isotope attributes.


	ArrayError
		if `plt_corr` is `True` but no corrected data exist.

	ArrayError
		If `result` does not contain any of: d13C, Fm.
	'''

	#extract the timedata dgdt
	try:
		dgdt = timedata.dgdt

	except AttributeError:
		#raise warning
		warnings.warn(
			'TimeData instance does not contain dgdt attribute! Proceeding'
			' to plot isotopes only!')

		#proceed with dgdt = None
		dgdt = None

	#extract t array also
	t = timedata.t

	#extract isotope data
	if plt_corr is True:
		flag = '_corr'
	
	else:
		flag = '_raw'

	#check that isotope attributes exist and extract if so
	d13C = getattr(result, 'd13C' + flag, None)
	d13C_std = getattr(result, 'd13C' + flag + '_std', None)

	Fm = getattr(result, 'Fm' + flag, None)
	Fm_std = getattr(result, 'Fm' + flag + '_std', None)

	#raise warning if no isotope data exist
	if d13C is None and Fm is None:
		#raise warning
		warnings.warn(
			'Result instance does not contain d13C or Fm attributes!'
			' Proceeding to thermogram only!')

	#create axis if necessary
	if ax is None:
		_, ax = plt.subplots(1, 1)

	#calculate time elapsed in each fraction and midpoints
	DT = result.t_frac[:,1] - result.t_frac[:,0]
	x = result.t_frac[:,0] + DT/2

	#create string to add variables to for legend
	lns = []

	#plot d13C
	if d13C is not None:
		d13C_plot = ax.errorbar(
			x,
			d13C,
			yerr = d13C_std,
			marker = 'o',
			ecolor = 'k',
			markersize = 12,
			mec = 'k',
			mfc = 'w',
			elinewidth = 1,
			markeredgewidth = 1,
			capsize = 0,
			ls = 'none',
			zorder = 3,
			label = r'fraction $\delta^{13}C$')

		#set label
		ax.set_ylabel(r'$\delta^{13}C$ (VPDB)')

		#append legend
		lns.append(d13C_plot)

	#copy axis
	ax2 = ax.twinx()

	#plot Fm, if it exists
	if Fm is not None:

		Fm_plot = ax2.bar(
			result.t_frac[:,0] + DT/2,
			Fm,
			width = DT,
			color = 'none',
			edgecolor = 'k', #ADDED 5 NOV 2018
			linewidth = 1.5,
			zorder = 2,
			label = 'fraction Fm')

		#set label
		ax2.set_ylabel('Fm')

		#append legend
		lns.append(Fm_plot)

		#find max Fm value
		mFm = np.max(Fm)
	else:
		mFm = 1.0

	#add thermogram plot if it exists
	if dgdt is not None:
		
		#scale thermogram to match Fm
		mtg = np.max(-dgdt)
		tg_scaled = -dgdt*mFm/mtg

		#plot thermogram
		tg_plot = ax2.fill_between(
			t,
			np.zeros(len(t)),
			tg_scaled,
			facecolor = [0.5, 0.5, 0.5],
			edgecolor = 'k',
			alpha = 0.3,
			label = 'thermogram (normalized)',
			zorder = 1)

		#append legend
		lns.append(tg_plot)

	#set x label and limits
	ax.set_xlabel('time (seconds)')
	ax2.set_ylim([0, 1.1*mFm])

	#get legend of everything in one
	labs = [l.get_label() for l in lns]
	
	ax.legend(
		lns,
		labs, 
		loc = 'best',
		frameon = False)

	#make tight layout
	plt.tight_layout()

	return ax, ax2




'''
This module contains helper functions for the Results class.
'''

from __future__ import(
	division,
	print_function,
	)

__docformat__ = 'restructuredtext en'
__all__ = ['_calc_cutoff', '_calc_E_frac','_rpo_blk_corr',
	'_rpo_extract_iso', '_rpo_kie_corr', '_rpo_mass_bal_corr']

import numpy as np
import pandas as pd
import warnings

from collections import Sequence
from numpy.linalg import norm
from scipy.optimize import least_squares
from scipy.optimize import nnls

#import Daem
from .model import(
	Daem
	)

#import exceptions
from .exceptions import(
	ArrayError,
	FileError,
	LengthError,
	ScalarError,
	)

#import helper functions
from .core_functions import(
	assert_len,
	extract_moments,
	)

#define a function to calculate cutoff indices for each RPO fraction
def _calc_cutoff(result, model):
	'''
	Calculates the time index corresponding to the start and stop of each RPO
	fraction. Used for calculating E_frac.

	Parameters
	----------
	result : rp.RpoIsotopes
		``RpoIsotopes`` instance containing CO2 fraction information.

	model : rp.Model
		``rp.Model`` instance containing the A matrix to use for inversion.

	Returns
	-------
	ind_min : np.ndarray
		Index in ``timedata.t`` corresponding to the minimum time for each 
		fraction. Length nFrac.

	ind_max : np.ndarray
		Index in ``timedata.t`` corresponding to the maximum time for each 
		fraction. Length nFrac.
	'''

	#extract shapes
	nFrac = result.nFrac
	nt = model.nt
	
	#extract arrays
	t_frac = result.t_frac
	t = model.t

	#pre-allocate index arrays
	ind_min = np.zeros(nFrac, dtype = int)
	ind_max = np.zeros(nFrac, dtype = int)

	#loop through and calculate indices
	for i, row in enumerate(t_frac):

		#extract indices for each fraction
		ind = np.where((t > row[0]) & (t <= row[1]))[0]

		#store first and last indices
		if ind[0] > 0:
			ind_min[i] = ind[0] - 1 #subtract one so theres no gap!
		else:
			ind_min[i] = ind[0] #don't subtract one if it makes it negative

		ind_max[i] = ind[-1]

	return ind_min, ind_max

#define function to calculate the E values of each RPO fraction
def _calc_E_frac(result, model, ratedata):
	'''
	Method for determining the distribution of E values contained within
	each RPO fraction. Used for calculating `E_frac` and `E_frac_std` in
	order to compare with measured isotope values.

	Parameters
	----------
	result : rp.RpoIsotopes
		``rp.RpoIsotopes`` instance containing the `t_frac` array to be used
		for calculating E in each fraction.

	model : rp.Model
		``rp.Model`` instance containing the A matrix to use for 
		inversion.

	ratedata : rp.RateData
		``rp.Ratedata`` instance containing the reactive continuum data.

	Returns
	-------
	E_frac : np.ndarray
		Array of mean E value contained in each RPO fraction, length `nFrac`.

	E_frac_std : np.ndarray
		Array of the standard deviation of E contained in each RPO fraction,
		length `nFrac`.

	p_frac : np.ndarray
		2d array of the distribution of E contained in each RPO fraction,
		shape [`nFrac` x `nE`].
	'''

	#calculate cutoff indices
	ind_min, ind_max = _calc_cutoff(result, model)

	#extract necessary data
	A = model.A
	E = ratedata.E
	dE = np.gradient(E)
	nE = ratedata.nE
	nF = result.nFrac
	p = ratedata.p

	#make an empty matrix to store results
	p_frac = np.zeros([nF, nE])
	E_frac = np.zeros(nF)
	E_frac_std = np.zeros(nF)

	#loop through each time window and calculate p0E_diff distribution
	for i in range(nF):

		#extract indices
		imin = ind_min[i]
		imax = ind_max[i]

		#p(E,t) at time 0
		pt0 = p*A[imin,:]/dE #divide by dE so total area == 1

		#p(E,t) at time final
		ptf = p*A[imax,:]/dE #divide by dE so total area == 1

		#difference -- i.e. p(E) evolved over Dt
		Dpt = pt0 - ptf

		#store in matrix
		p_frac[i, :] = Dpt

		#calculate the mean and stdev
		E_frac[i], E_frac_std[i] = extract_moments(E, Dpt)

	return E_frac, E_frac_std, p_frac

#define a function to blank-correct fraction isotopes
def _rpo_blk_corr(
		d13C, 
		d13C_std, 
		Fm, 
		Fm_std, 
		m, 
		m_std, 
		t,
		blk_d13C = (-29.0, 0.1),
		blk_flux = (0.375, 0.0583),
		blk_Fm =  (0.555, 0.042)):
	'''
	Performs blank correction on raw isotope values.

	Parameters
	----------
	d13C : None or np.ndarray
		Array of d13C values for each fraction, length nFrac.
	
	d13C_std : np.ndarray
		Array of d13C stdev. for each fraction, length nFrac.

	Fm : None or np.ndarray
		Array of Fm values for each fraction, length nFrac.

	Fm_std : np.ndarray
		Array of Fm stdev. for each fraction, length nFrac.

	m : np.ndarray
		Array of masses (ugC) for each fraction, length nFrac.

	m_std : np.ndarray
		Array of mass stdev. (ugC) for each fraction, length nFrac.

	t : np.ndarray
		2d array of time for each fraction (in seconds), length nFrac.

	blk_d13C : tuple
		Tuple of the blank d13C composition (VPDB), in the form 
		(mean, stdev.) to be used of ``blk_corr = True``. Defaults to the
		NOSAMS RPO blank as calculated by Hemingway et al. **2017**.

	blk_flux : tuple
		Tuple of the blank flux (ng/s), in the form (mean, stdev.) to
		be used of ``blk_corr = True``. Defaults to the NOSAMS RPO blank 
		as calculated by Hemingway et al. **2017**.

	blk_Fm : tuple
		Tuple of the blank Fm value, in the form (mean, stdev.) to
		be used of ``blk_corr = True``. Defaults to the NOSAMS RPO blank 
		as calculated by Hemingway et al. **2017**.

	Returns
	-------
	d13C_corr : None or np.ndarray
		Array of corrected d13C values for each fraction, length nFrac.
	
	d13C_std_corr : np.ndarray
		Array of corrected d13C stdev. for each fraction, length nFrac.

	Fm_corr : None or np.ndarray
		Array of corrected Fm values for each fraction, length nFrac.

	Fm_std_corr : np.ndarray
		Array of corrected Fm stdev. for each fraction, length nFrac.

	m_corr : np.ndarray
		Array of corrected masses (ugC) for each fraction, length nFrac.

	m_std_corr : np.ndarray
		Array of corrected mass stdev. (ugC) for each fraction, length nFrac.
	
	References
	----------
	[1] J.D. Hemingway et al. (2017) Assessing the blank carbon contribution, 
		isotope mass balance, and kinetic isotope fractionation of the ramped 
		pyrolysis/oxidation instrument at NOSAMS. **Radiocarbon**
	'''

	#define constants
	bl_flux = blk_flux[0]/1000 #make ug/s
	bl_flux_std = blk_flux[1]/1000 #make ug/s

	bl_d13C = blk_d13C[0]
	bl_d13C_std = blk_d13C[1]

	bl_Fm = blk_Fm[0]
	bl_Fm_std = blk_Fm[1]

	#calculate blank mass for each fraction
	dt = t[:,1] - t[:,0]
	bl_mass = bl_flux*dt #ug

	#perform blank correction

	#correct mass
	m_corr = m - bl_mass
	m_std_corr = norm(
		[m_std, dt*bl_flux_std], 
		axis = 0)

	#correct d13C
	if d13C is not None:

		dt1 = d13C_std
		dt2 = dt*bl_d13C*bl_flux_std/m
		dt3 = bl_mass*bl_d13C_std/m
		dt4 = bl_mass*bl_d13C*m_std_corr/(m_corr**2)
	
		d13C_corr = (m*d13C - bl_mass*bl_d13C)/m_corr
		d13C_std_corr = norm(
			[dt1, dt2, dt3, dt4], 
			axis = 0)

	else:
		d13C_corr = None
		d13C_corr_std = 0

	#correct Fm
	if Fm is not None:

		ft1 = Fm_std
		ft2 = dt*bl_Fm*bl_flux_std/m
		ft3 = bl_mass*bl_Fm_std/m
		ft4 = bl_mass*bl_Fm*m_std_corr/(m_corr**2)
		
		Fm_corr = (m*Fm - bl_mass*bl_Fm)/m_corr
		Fm_std_corr = norm(
			[ft1, ft2, ft3, ft4], 
			axis = 0)

	else:
		Fm_corr = None
		Fm_corr_std = 0

	return (d13C_corr, 
			d13C_std_corr, 
			Fm_corr, 
			Fm_std_corr, 
			m_corr, 
			m_std_corr)

#define function to extract Rpo isotope data from .csv file
def _rpo_extract_iso(file, mass_err):
	'''
	Extracts mass, d13C, and Fm from a .csv file to be used for to create an
	``RpoIsotopes`` result instance.

	Parameters
	----------
	file : str or pd.DataFrame
		File containing isotope data, either as a path string or 
		``pd.DataFrame`` instance.

	mass_err : float
		Relative standard deviation on fraction masses. Defaults to 0.01 (i.e.
		1 percent of measured mass).

	Returns
	-------
	d13C : None or np.ndarray
		Array of d13C values for each fraction, length nFrac.
	
	d13C_std : np.ndarray
		Array of d13C stdev. for each fraction, length nFrac.

	Fm : None or np.ndarray
		Array of Fm values for each fraction, length nFrac.

	Fm_std : np.ndarray
		Array of Fm stdev. for each fraction, length nFrac.

	m : None or np.ndarray
		Array of masses (ugC) for each fraction, length nFrac.

	m_std : np.ndarray
		Array of mass stdev. (ugC) for each fraction, length nFrac.

	t : np.ndarray
		2d array of time for each fraction (in seconds), length nFrac.

	Raises
	------
	FileError
		If `file` does not contain "fraction" column.

	FileError
		If `file` is not str or ``pd.DataFrame``.
	
	FileError
		If index is not ``pd.DatetimeIndex`` instance.	
	
	FileError
		If first two rows are not fractions "-1" and "0"

	ScalarError
		If `mass_err` is not scalar.
	
	Notes
	-----
	For bookkeeping purposes, the first 2 rows must be fractions "-1" and "0",
	where the timestamp for fraction "-1" is the first point in `all_data` and
	the timestamp for fraction "0" is the t0 for the first fraction.
	'''

	#import file as a pd.DataFrame if inputted as a string path and check
	#that it is in the right format
	if isinstance(file, str):
		
# REVISION NOTES
# 2023.05.29: CHANGED 'parse_dates = True' TO PARSE AS OBJECT AND THEN
#			  PARSE DIRECTLY USING 'to_datetime()'. FIXED WARNING.

		#import as dataframe
		file = pd.read_csv(
			file,
			index_col = 0,
			# parse_dates = True
			parse_dates = False
			)

		#convert index to datetime
		ind = file.index.astype(object)
		ind_dt = pd.to_datetime(ind,format='%I:%M:%S %p')

		#now re-save as index
		file.index = ind_dt

	elif not isinstance(file, pd.DataFrame):
		raise FileError(
			'file must be pd.DataFrame or path string')

	if 'fraction' not in file.columns:
		raise FileError(
			'file must have "fraction" column')

	if not isinstance(file.index, pd.DatetimeIndex):
		raise FileError(
			'file index must be DatetimeIndex')

	if file.fraction[0] != -1 or file.fraction[1] != 0:
		raise FileError(
			'First two rows must be fractions "-1" and "0"')

	if not isinstance(mass_err, (str, float)):
		raise ScalarError(
			'mass_err must be string or float')
	else:
		#ensure float
		mass_err = float(mass_err)

	#extract time data
	secs = (file.index - file.index[0]).seconds
	t0 = secs[1:-1]
	tf = secs[2:]
	nF = len(t0)

	t = np.column_stack((t0, tf))

	#extract mass and isotope data, if they exist
	if 'ug_frac' in file.columns:
		m = file.ug_frac[2:].values
		m_std = m*mass_err

	else:
		m = None
		m_std = None

	if 'd13C' in file.columns:
		d13C = file.d13C[2:].values
		d13C_std = file.d13C_std[2:].values

	else:
		d13C = None
		d13C_std = None

	if 'Fm' in file.columns:
		Fm = file.Fm[2:].values
		Fm_std = file.Fm_std[2:].values

	else:
		Fm = None
		Fm_std = None

	return (d13C,
			d13C_std,
			Fm,
			Fm_std,
			m,
			m_std,
			t)

#define function to correct d13C for kinetic fractionation
def _rpo_kie_corr(
	result,
	d13C,
	d13C_std,
	model,
	ratedata,
	DE = 0.0018):

	'''
	Corrects d13C values for each RPO fraction for kinetic isotope effects.

	Parameters
	----------
	result : rp.Results
		Result instance containing the start/stop times to be used to calculate
		E_frac

	d13C : np.ndarray
		Array of d13C values to correct.

	d13C_std : np.ndarray
		Array of the standard deviation of d13C values to correct.

	model : rp.Model
		``rp.Model`` instance containing the A matrix to use for inversion.

	ratedata : rp.RateData
		``rp.Ratedata`` instance containing the reactive continuum data.

	DE : scalar
		Value for the difference in E between 12C- and 13C-containing
		atoms, in kJ. Defaults to 0.0018 (the best-fit value calculated
		in Hemingway et al., **2017**).

	Returns
	-------
	d13C_corr : np.ndarray
		Array of the fractionation-corrected d13C values (VPDB) of each 
		measured fraction, length `nFrac`.

	d13C_corr_std : np.ndarray
		The standard deviation of `d13C_corr` with length `nFrac`.
	'''

	#generate daem for 13C-containing atoms
	daem13 = Daem(model.E+DE, model.log10omega, model.t, model.T)

	#calculate ratio of rates, the KIE
	tg12 = np.dot(model.A,ratedata.p)
	tg13 = np.dot(daem13.A,ratedata.p)
	r = np.gradient(tg13)/np.gradient(tg12)

	#for each time slice, calculate KIE-corrected d13C
	d13C_corr = []

	#calculate cutoff indices
	ind_min, ind_max = _calc_cutoff(result, model)

	#loop through and correct each measurement
	for mi, ma, d13i in zip(ind_min, ind_max, d13C):
		
		#weighted-average kie for each slice
		kie_i = np.average(r[mi:ma], weights = -np.gradient(tg12)[mi:ma])

		#convert measured d13C to ratio
		R13_i = (d13i/1000 + 1)*0.011237

		#divide by kie_i to correct
		R13_i_corr = R13_i/kie_i

		#convert back to d13C
		d13C_i_corr = (R13_i_corr/0.011237 - 1)*1000

		#append list
		d13C_corr.append(d13C_i_corr)

	#since uncertainty is unknown, d13C_std is unchanged
	d13C_corr_std = d13C_std

	return d13C_corr, d13C_corr_std

#define function to correct d13C for isotope mass balance
def _rpo_mass_bal_corr(
	d13C,
	d13C_std,
	m,
	m_std,
	bulk_d13C_true):
	'''
	Corrects d13C values for the difference between mass-weighted mean and
	independently measured bulk values (i.e. isotope mass balance).

	Parameters
	----------
	d13C : np.ndarray
		Array of the d13C values (VPDB) of each measured fraction, 
		length `nFrac`.

	d13C_std : np.ndarray
		The standard deviation of `d13C` with length `nFrac`.

	m : np.ndarray
		Array of the masses (ugC) of each measured fraction, length `nFrac`.

	m_frac : np.ndarray
		The standard deviation of `m_frac` with length `nFrac`.

	bulk_d13C_true: array
		The true, independently measured bulk d13C value for the sample.
		Inputted in the form [mean, stdev.].

	Returns
	-------
	d13C_corr : np.ndarray
		Array of the mass-balance-corrected d13C values (VPDB) of each
		measured fraction, length `nFrac`.

	d13C_corr_std : np.ndarray
		The standard deviation of `d13C_corr` with length `nFrac`.

	Raises
	------
	ArrayError
		If `bulk_d13C_true` is not array-like

	LengthError
		If `bulk_d13C_true` is not an array of length 2.

	References
	----------
	[1] J.D. Hemingway et al. (2017) Assessing the blank carbon contribution, 
		isotope mass balance, and kinetic isotope fractionation of the ramped 
		pyrolysis/oxidation instrument at NOSAMS. **Radiocarbon**
	'''

	#ensure bulk_d13C_true is in the right format, and raise error if not
	if not isinstance(bulk_d13C_true, Sequence) and not hasattr(bulk_d13C_true, '__array__'):
		raise ArrayError(
			'bulk_d13C_true must be array-like')

	if len(bulk_d13C_true) != 2:
		raise LengthError(
			'length of bulk_d13C_true must be 2 [mean, stdev.]')

	#calculate the fractional contribution by each RPO fraction
	f = m/np.sum(m)

	#calculate the weighted-average d13C, d13C_std
	d13C_wgh = np.average(d13C, weights = f)
	d13C_wgh_std = norm(f*d13C_std)

	#calculate mass-balance-corrected values
	d13C_corr = d13C + (bulk_d13C_true[0] - d13C_wgh)
	d13C_corr_std = (d13C_std**2 + bulk_d13C_true[1]**2 + d13C_wgh_std**2)**0.5

	return d13C_corr, d13C_corr_std

